#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <assert.h>
#include "../lib/misc.h"
#include "../lib/struct.h"

double meaScore(char* secStr, int alpha, int beta, double** P, double* Q) {
	list* l = getBasePairList(secStr);
	double val = 0;
	int* position = (int*) malloc(strlen(secStr) * sizeof(int));
	int i;
	for (i = 0; i < strlen(secStr); i++) {
		position[i] = 1;
	}
	item* parser = l->first;
	while (parser != NULL) {
		val += 2 * alpha * P[parser->x][parser->y];
		position[parser->x] = 0;
		position[parser->y] = 0;
		parser = parser->next;
	}
	for (i = 0; i < strlen(secStr); i++) {
		if (position[i] == 1)
			val += beta * Q[i];
	}
	free(position);
	free_list(l);
	return val;
}
void initializeMatrix(char *rna, char *S0, int Kmax, double **P, double *Q,
		double ***M, double ****N, int alpha, int beta, char *constraint) {
	/*
	 * M is dictionary of form M[(i,j,k)] = MEA score of rna[i],...,rna[j]
	 * with respect to structures having k base pair difference with S0
	 * This function initializes energy to be 0 for k>0
	 * WARNING: S0 is unchanged, but secStrList0 is list of (i,j) with 1..i,j..n
	 * Indices of RNA go from 1,...,n due to dummy character '$'
	 */
	list* secStrList0 = getBasePairList(S0);
	int n = strlen(rna);
	int i, j, k, d, r;
	double sum;
	// now set to correct values
	for (d = 0; d < n; d++) { // d = 0, ... , n-1
		for (i = 0; i < n - d; i++) { // i = 0, ... , n-d-1
			j = i + d; // j = 0, ... , n-1
			if (j <= i + THRESHOLD) { // i<=j<=i+THRESHOLD
				sum = 0;
				int check = 1;
				for (r = i; r <= j; r++) {
					if (constraint[r] == '(' || constraint[r] == ')')
						check = 0;
				}
				if (check == 1) {
					for (r = i; r <= j; r++) {
						sum = sum + beta * Q[r];
					}
					M[i][j][0] = sum;
				}
				// N[(i,j,0)] already set to (0,0,0)
			} else if (i + THRESHOLD + 1 == j) {
				double val = 0;
				for (r = i + 1; r < j; r++) { // val = beta*(Q[i+1]+...+Q[j-1])
					val += beta * Q[r];
				}
				if (existItemInList(secStrList0, i, j) == 1) {
					if ((constraint[i] == '(' && constraint[j] == ')')
							|| (constraint[i] == '.' && constraint[j] == '.')) {
						M[i][j][0] = 2 * alpha * P[i][j] + val;
						N[i][j][0][0] = i;
					}
					if ((constraint[i] == '*' || constraint[i] == '.')
							&& (constraint[j] == '*' || constraint[j] == '.')) {
						M[i][j][1] = val + beta * (Q[i] + Q[j]);
					}
					// N[(i,j,1)] already set to (0,0,0)
				} else { // (i,j) not in secStrList0
					M[i][j][0] = val + beta * (Q[i] + Q[j]);
					// N[(i,j,0)] already set to (0,0,0)
					if (basePair(rna[i], rna[j])) {
						if ((constraint[i] == '(' && constraint[j] == ')')
								|| (constraint[i] == '.' && constraint[j]
										== '.')) {
							M[i][j][1] = 2 * alpha * P[i][j] + val;
							N[i][j][1][0] = i;
						}
					}
				}
			}
		}
	}
	free_list(secStrList0);
}
void computeEnergyMatrix(char* rna, char* S0, int Kmax, double ***M,
		double **** N, double **P, double *Q, int alpha, int beta,
		char *constraint) {
	/*
	 *  P is dictionary of base pairing probability, P[i][j] defined for i<j
	 *  P[i][j] = p, where p is probability of (i,j), 1<=i<j<=n
	 *  Q[i]    = 1 - sum_{j < i} P[j][i] - sum_{i<j} P[i][j], prob of unpaired i
	 *  paren is parenthesis expression for maximal RNA secondary structure
	 *  Function computes the maximum expected accurate sec str
	 *  with MEA score 2* alpha * sum P[i][j] + beta * Q[i] over sec str with
	 *  base pair distance k from S0
	 *
	 * Function returns secStr,numBP, where secStr of form (((...)))
	 * NOTE: secStr has NO leading dummy char '$'
	 */
	int n = strlen(rna);
	list *secStrList0 = getBasePairList(S0);
	int i, j, k, k0, k1, d, r, b0, b1, b2;
	double max, val;
	double *index;
	for (d = THRESHOLD + 2; d < n; d++) {
		for (i = 0; i < n - d; i++) {
			for (k = 0; k <= Kmax; k++) {
				j = i + d;
				max = -INF;
				index = (double *) malloc(3 * sizeof(double));
				index[0] = -1;
				index[1] = 0;
				index[2] = 0;
				// Set M[(i,j,k)] to maximum of following cases
				// Case 1: j unpaired in S[i,j], where S in NOT S0, but arbitrary str
				if (constraint[j] == '*' || constraint[j] == '.') {
					b0 = case1(secStrList0, i, j);
					if (k - b0 >= 0) {
						val = M[i][j - 1][k - b0] + beta * Q[j];
						if (val > max) {
							max = val;
							index[1] = k - b0;
							// backtracking: j unpaired, k-b0 bpdistance on remainder
						}
					}
				}
				//Case 2: (i,j) in S, where S is NOT S0, but arbitrary str
				if ((constraint[i] == '(' && constraint[j] == ')')
						|| (constraint[i] == '.' && constraint[j] == '.')) {
					if (basePair(rna[i], rna[j])) { // check if i,j can pair
						if ((constraint[i] == '(' && constraint[j] == ')')
								|| (constraint[i] == '.' && constraint[j]
										== '.')) {
							b1 = case2(secStrList0, i, j);
							//b1 = bpDist(S0[i+1,j-1] +{(i,j)},S0[i,j])
							if (k - b1 >= 0) {
								val = M[i + 1][j - 1][k - b1] + 2 * alpha
										* P[i][j];
								if (val > max) {
									max = val;
									index[0] = i;
									index[2] = k - b1; // backtracking: (i,j) in S
								}
							}
						}
					}
				}
				// Case 3: (r,j) in S, some i<r<j, where S is NOT S0, but arbitrary str
				for (r = i + 1; r < j - THRESHOLD; r++) {
					if (basePair(rna[r], rna[j])) {
						if ((constraint[r] == '(' && constraint[j] == ')')
								|| (constraint[r] == '.' && constraint[j]
										== '.')) {
							b2 = case3(secStrList0, i, j, r);
							// b2 = bpDist(S0[i,r-1]+S0[r+1,j-1]+{(r,j)},S0[i,j])
							for (k0 = 0; k0 < k - b2 + 1; k0++) {
								k1 = k - b2 - k0; // k0+k1+b2 = k
								if (k1 >= 0 && M[i][r - 1][k0] > 0 && M[r + 1][j - 1][k1] > 0) {
									val = M[i][r - 1][k0] + M[r + 1][j - 1][k1]
											+ 2 * alpha * P[r][j];
									if (val > max) {
										max = val;
										index[0] = r;
										index[1] = k0;
										index[2] = k1;
										// backtracking: (r,j) in S, k0 bpDist in left, k1 bpDist in right
									}
								}
							}
						}
					}
				}
				M[i][j][k] = max;
				N[i][j][k][0] = index[0];
				N[i][j][k][1] = index[1];
				N[i][j][k][2] = index[2];
				free(index);
			}
		}
	}
	free_list(secStrList0);
}
void printMatrix(double ***M, int n, int Kmax) {
	printf("Matrix values");
	int i, j, k;
	double z;
	for (k = 0; k < Kmax + 1; k++) {
		printf(" k=%d ", k);
	}
	for (i = 0; i < n; i++) {
		for (j = 0; j < n; j++) {
			z = M[i][j][k];
			if (i <= j) {
				printf("%.2f ", z);
			}
		}
	}
	printf("\n");
}
void computeMEA_RNAbor(char *rna, char *S0, int Kmax, double **P, double *Q,
		int alpha, int beta, char *id, char *constraint) {
	/*
	 *  P is dictionary of base pairing probability, P[i][j] defined for i<j
	 *  P[i][j] = p, where p is probability of (i,j), 1<=i<j<=n
	 *  Q[i] = 1 - sum_{j < i} P[j][i] - sum_{i<j} P[i][j], prob of unpaired i
	 *  paren is parenthesis expression for maximal RNA secondary structure
	 *  Function computes the maximum expected accurate sec str
	 *  with MEA score alpha  * sum P[i][j] + beta * sum Q[i] over sec str with
	 *  base pair distance k from S0
	 *
	 * Function returns secStr,numBP, where secStr of form (((...)))
	 * NOTE: secStr has NO leading dummy char '$'
	 */
	int i, j, k;
	int n = strlen(rna);
	/*----- initialization  --*/
	double ***M = (double ***) malloc(n * sizeof(double **));
	for (i = 0; i < n; i++) {
		M[i] = (double **) malloc(n * sizeof(double *));
		for (j = 0; j < n; j++) {
			M[i][j] = (double *) malloc((Kmax + 1) * sizeof(double));
			for (k = 0; k <= Kmax; k++) {
				M[i][j][k] = -INF;
			}
		}
	}
	double ****N = (double ****) malloc(n * sizeof(double ***));
	for (i = 0; i < n; i++) {
		N[i] = (double ***) malloc(n * sizeof(double **));
		for (j = 0; j < n; j++) {
			N[i][j] = (double **) malloc((Kmax + 1) * sizeof(double *));
			for (k = 0; k <= Kmax; k++) {
				N[i][j][k] = (double *) malloc(3 * sizeof(double));
				N[i][j][k][0] = -1;
				N[i][j][k][1] = 0;
				N[i][j][k][2] = 0;
			}
		}
	}
	initializeMatrix(rna, S0, Kmax, P, Q, M, N, alpha, beta, constraint);
	computeEnergyMatrix(rna, S0, Kmax, M, N, P, Q, alpha, beta, constraint);
	if (DEBUG) {
		for (i = 0; i < n; i++) {
			for (j = 0; j < n; j++) {
				for (k = 0; k < n; k++) {
					printf("%d %d %d, %f", i, j, k, M[i][j][k]);
				}
			}
		}
	}
	/* write info into file 'out.rnabor'*/
	int numSecStr = 0;
	char leftBracket = '(';
	char rightBracket = ')';
	int dist;
	double mea;
	double energy;
	char filename[256];
	sprintf(filename, "out/%s.rnabor", id);
	FILE *fo = fopen(filename, "w");
	fprintf(fo, "Distance\tMEA\tStructure\tEnergy\n");
	char *paren = (char *) malloc((n + 1) * sizeof(char));
	for (k = 0; k <= Kmax; k++) {
		for (i = 0; i < n; i++) {
			paren[i] = '.';
		}
		paren[i] = '\0';
		if (M[0][n - 1][k] > 0) {
			if (DEBUG)
				printf("BACKTRACK %d\n", k);
			backtrack(0, n - 1, k, M, N, paren, leftBracket, rightBracket);
			numSecStr++;
			dist = basePiarDisntance(S0, paren);
			assert(dist == k);
			mea = meaScore(paren, alpha, beta, P, Q);
			assert(fabs(mea - M[0][n - 1][k]) < 0.00001);
			energy = evalViennaEnergy(rna, paren);
			fprintf(fo, "%d \t %f \t %s %f\n", k, M[0][n - 1][k], paren, energy);
		}
	}
	fclose(fo);
	/* free */
	for (i = 0; i < n; i++) {
		for (j = 0; j < n; j++) {
			for (k = 0; k <= Kmax; k++) {
				free(N[i][j][k]);
			}
			free(N[i][j]);
			free(M[i][j]);
		}
		free(N[i]);
		free(M[i]);
	}
	free(M);
	free(N);
	free(paren);
}
void backtrack(int i, int j, int k, double ***M, double ****N, char *paren,
		char leftBracket, char rightBracket) {
	// backtrack on [i,j] for k-nbor of S0 restricted to [i,j]
	if (j - i > THRESHOLD && M[i][j][k] > 0) {
		if (DEBUG) {
			printf("%d %d %d %f %d %d %d", i, j, k, M[i][j][k], N[i][j][k][0],
					N[i][j][k][1], N[i][j][k][2]);
		}
		int r = N[i][j][k][0];
		int k0 = N[i][j][k][1];
		int k1 = N[i][j][k][2];
		// j base-pairs with r in [i,j]
		if (r >= 0) {
			paren[r] = leftBracket; // paren has dummy char at position 0
			paren[j] = rightBracket; // so indices of paren and energy go from 1,...,n
			if (DEBUG) {
				printf("+++1) backtrack(%d, %d, %d, %s)", r + 1, j - 1, k1,
						paren);
			}
			backtrack(r + 1, j - 1, k1, M, N, paren, leftBracket, rightBracket);
			if (DEBUG) {
				printf("+++2) backtrack(%d,%d,%d,%s)", i, r - 1, k0, paren);
			}
			backtrack(i, r - 1, k0, M, N, paren, leftBracket, rightBracket);
		} else {
			// r is 0, so j not paired in [i,j]
			backtrack(i, j - 1, k0, M, N, paren, leftBracket, rightBracket);
		}
	}
}
